@@ -3,18 +3,20 @@ import marshal
 import sys
 import subprocess
 import socket
+from pprint import pformat
 
 import logging
 
 from . import p4errors
 from .p4file import P4File, Status
+from .utils import split_list_into_strings_of_length, convert_to_list, decode_dictionaries, validate_not_empty
 
 MAX_CMD_LEN = 8190
 MAX_ARG_LEN = 8000  # max length of args string when combined, close to max, but leaving some extra margin
 
 
 class P4Client(object):
-    def __init__(self, perforce_root, user=None, client=None, server=None, silent=True):
+    def __init__(self, perforce_root, user=None, client=None, server=None, silent=True, max_parallel_connections=4):
         """
         Make a new P4Client
 
@@ -22,10 +24,13 @@ class P4Client(object):
         :param user: *string* P4USER, if None will be tried to be found automatically
         :param client: *string* P4CLIENT, if None will be tried to be found automatically
         :param silent: *bool* if True, suppresses error messages to cut down on terminal spam
+        :param max_parallel_connections: *int* max number of connections to use while syncing/submitting. This requires
+        the server to have net.parallel.max and net parallel.threads to be > 1
         """
         self.perforce_root = perforce_root
         self.silent = silent
 
+        self.max_parallel_connections = max_parallel_connections
         self.user = user
         self.client = client
         self.server = server
@@ -47,10 +52,6 @@ class P4Client(object):
             if self.server is None:
                 raise p4errors.WorkSpaceError("Could not find P4PORT")
 
-
-        # print(self.user)
-        # print(self.client)
-
     @classmethod
     def from_env(cls, *args, **kwargs):
         """
@@ -70,6 +71,15 @@ class P4Client(object):
         """
         self.perforce_root = root
 
+    def set_max_parallel_connections(self, value):
+        """
+        Set the number of maximum parallel connections to use for sync/submit. Default class value is 4
+
+        :param value: *int*
+        :return:
+        """
+        self.max_parallel_connections = value
+
     def run_cmd(self, cmd, args=[], file_list=[], use_global_options=True, online_check=True):
         """
         Reads the output stream of the command and returns it as a marshaled dict.
@@ -99,7 +109,6 @@ class P4Client(object):
         for clamped_arg in clamped_arg_list:
             for clamped_files in clamped_file_list:
                 if use_global_options:
-                    # command = f"p4 -G {' '.join(global_options)} {cmd} {clamped_arg}"
                     command = f"p4 -G -u {self.user} -c {self.client} {cmd} {clamped_arg} {clamped_files}"
                 else:
                     command = f"p4 {cmd} {clamped_arg} {clamped_files}"
@@ -108,24 +117,23 @@ class P4Client(object):
                     # This shouldn't happen, but just in case the command prefix end up really long
                     logging.warning(f"Command length: {format(len(command))} exceeds MAX_CMD_LEN {MAX_CMD_LEN} on command: {MAX_CMD_LEN}")
 
-                # print(command)
-                pipe = subprocess.Popen(command, stdin=subprocess.PIPE, stdout=subprocess.PIPE, shell=True)
-                output = pipe.stdout
-
-                try:
-                    while True:
-                        value_dict = marshal.load(output)
-                        dict_list.append(value_dict)
-                except EOFError:
-                    pass
-                except ValueError as error:
-                    output_dict = {
-                        "command": command,
-                        "code": "error",
-                        "error": str(error),
-                        "raw_output": subprocess.check_output(command, stderr=subprocess.STDOUT, shell=True)
-                    }
-                    dict_list.append(output_dict)
+                with subprocess.Popen(command, stdin=subprocess.PIPE, stdout=subprocess.PIPE, shell=True) as pipe:
+                    output = pipe.stdout
+                    try:
+                        while True:
+                            value_dict = marshal.load(output)
+                            dict_list.append(value_dict)
+                    except EOFError:
+                        pass
+                    except ValueError as error:
+                        output_dict = {
+                            "command": command,
+                            "code": "error",
+                            "error": str(error),
+                            "raw_output": subprocess.check_output(command, stderr=subprocess.STDOUT, shell=True)
+                        }
+                        dict_list.append(output_dict)
+                    pipe.kill()
 
         return dict_list
 
@@ -199,6 +207,7 @@ class P4Client(object):
         else:
             raise p4errors.WorkSpaceError("Tried to set a workspace/client({}) that did not exist".format(workspace))
 
+    @validate_not_empty
     def files_to_p4files(self, file_list, allow_invalid_files=False):
         """
         Turns a list of files into P4File objects. If the Perforce server can't be reached, returns a list of P4Files
@@ -213,7 +222,7 @@ class P4Client(object):
 
         if self.host_online():
             fstat_output = self.run_cmd("fstat", file_list=file_list)
-            p4files = self.__fstat_to_p4_files(fstat_output, allow_invalid_files=allow_invalid_files)
+            p4files = self.fstat_to_p4_files(fstat_output, allow_invalid_files=allow_invalid_files)
             return p4files
         else:
             p4files = []
@@ -225,6 +234,7 @@ class P4Client(object):
                 p4files.append(perforce_file)
             return p4files
 
+    @validate_not_empty
     def folder_to_p4files(self, folder, include_subfolders=True, allow_invalid_files=False):
         """
         Returns all the files in a folder as a list of P4File objects. Uses the files_to_p4files function if the host
@@ -244,7 +254,7 @@ class P4Client(object):
                 folder = folder + "*" if folder.endswith("/") or folder.endswith("\\") else folder + "/*"
 
             fstat_output = self.run_cmd("fstat", file_list=[folder])
-            p4files = self.__fstat_to_p4_files(fstat_output, allow_invalid_files=allow_invalid_files)
+            p4files = self.fstat_to_p4_files(fstat_output, allow_invalid_files=allow_invalid_files)
             return p4files
 
         else:
@@ -294,7 +304,8 @@ class P4Client(object):
             if len(changelists):
                 return True
             return False
-
+    
+    @validate_not_empty
     def move_files_to_changelist(self, file_list, changelist="default"):
         """
         Moves the files in file_list to a changelist. Makes a new changelist if the given one doesn't exist.
@@ -309,10 +320,29 @@ class P4Client(object):
 
         for info_dict in info_dicts:
             if self.__get_dict_value(info_dict, "code") == "error" and not self.silent:
-                print(self.__get_dict_value(info_dict, "data"))
+                logging.error(self.__get_dict_value(info_dict, "data"))
 
         return info_dicts
 
+    def combine_changelists(self, source_changelists, target_changelist):
+        """
+        Moves all the files from the source_changelists to the target_changelist
+
+        :param source_changelists: *list* source changelists that will be emptied
+        :param target_changelist: *string* or *int* changelist that will hold all the files
+        empty after moving the files
+        :return: *list* of info dictionaries
+        """
+        target_changelist = self.__ensure_changelist(target_changelist)
+        files_to_move = []
+        for source_cl in source_changelists:
+            source_cl = self.__ensure_changelist(source_cl)
+            files_to_move.extend(self.get_files_in_changelist(source_cl))
+
+        result = self.move_files_to_changelist(files_to_move, target_changelist)
+        return result
+
+
     def rename_file(self, old_file_path, new_file_path, changelist="default"):
         """
         P4 move-renames a file
@@ -345,6 +375,7 @@ class P4Client(object):
             return True
         return False
 
+    @validate_not_empty
     def revert_files(self, file_list, unchanged_only=False):
         """
         Reverts files in file_list
@@ -363,11 +394,14 @@ class P4Client(object):
 
         return info_dicts
 
-    def revert_folders(self, folder_list):
+
+    @validate_not_empty
+    def revert_folders(self, folder_list, unchanged_only=False):
         """
         Recursively reverts complete folders
 
         :param folder_list: *list* folder paths
+        :param unchanged_only: *bool*
         :return: *list* of info dicts
         """
         folder_list = convert_to_list(folder_list) if not isinstance(folder_list, list) else folder_list
@@ -381,7 +415,10 @@ class P4Client(object):
             folder += "/..."
             cleaned_folder_list.append(folder)
 
-        info_dicts = self.run_cmd("revert", args=[], file_list=cleaned_folder_list)
+        if unchanged_only:
+            info_dicts = self.run_cmd("revert", ["-a"], file_list=cleaned_folder_list)
+        else:
+            info_dicts = self.run_cmd("revert", args=[], file_list=cleaned_folder_list)
         return info_dicts
 
     def revert_changelist(self, unchanged_only=False, changelist="default"):
@@ -390,9 +427,29 @@ class P4Client(object):
         :param changelist: string or int value
         :param unchanged_only: *bool*
         """
-        files = self.get_files_in_changelist(changelist)
-        self.revert_files(files, unchanged_only=unchanged_only)
+        changelist = self.__ensure_changelist(changelist)
+        if unchanged_only:
+            return self.run_cmd("revert", args=["-a", "-c", changelist])
+        else:
+            return self.run_cmd("revert", args=["-c", changelist, "//..."])
 
+    def submit_changelist(self, changelist, revert_unchanged_files=True):
+        """
+        Submits a changelist
+
+        :param changelist: string or int value.
+        :param revert_unchanged_files: *bool* does what it says on the box
+        :return:
+        """
+        changelist = self.__ensure_changelist(changelist)
+
+        if revert_unchanged_files:
+            self.revert_changelist(unchanged_only=True, changelist=changelist)
+
+        info_dicts = self.run_cmd("submit", args=["-c", changelist, "--parallel", f"threads={self.max_parallel_connections}"])
+        return info_dicts
+
+    @validate_not_empty
     def sync_folders(self, folder_list):
         """
         Recursively syncs complete folders
@@ -411,9 +468,10 @@ class P4Client(object):
             folder += "/..."
             cleaned_folder_list.append(folder)
 
-        info_dicts = self.run_cmd("sync", args=[], file_list=cleaned_folder_list)
+        info_dicts = self.run_cmd("sync", args=["--parallel", f"threads={self.max_parallel_connections}"], file_list=cleaned_folder_list)
         return info_dicts
 
+    @validate_not_empty
     def sync_files(self, file_list, revision=-1, verify=True, force=False):
         """
         Syncs files
@@ -430,7 +488,7 @@ class P4Client(object):
             verify = False
             file_list = [f"{path}#{revision}" for path in file_list]
 
-        initial_arg_list = ["-f"] if force else []
+        initial_arg_list = ["-f", "--parallel", f"threads={self.max_parallel_connections}"] if force else ["--parallel", f"threads={self.max_parallel_connections}"]
         if not self.silent:
             self.__validate_file_list(file_list)
 
@@ -444,6 +502,48 @@ class P4Client(object):
 
         return info_dicts
 
+    @validate_not_empty
+    def reconcile_offline_files(self, file_list, changelist="default"):
+        """
+        Adds, opens for edit or delete any files that were changed outside Perforce
+
+        :param file_list: *list* of local files
+        :param changelist: string or int value
+        :return: *list* of info dicts
+        """
+        file_list = convert_to_list(file_list) if not isinstance(file_list, list) else file_list
+
+        changelist = self.__ensure_changelist(changelist)
+
+        info_dicts = self.run_cmd("reconcile", args=["-c", changelist], file_list=file_list)
+        return info_dicts
+
+    @validate_not_empty
+    def reconcile_offline_folders(self, folder_list, changelist="default"):
+        """
+        Adds, opens for edit or delete any files in the rootfolder + subfolders that were changed outside Perforce
+
+        :param folder_list: *list* folders
+        :param changelist: string or int value
+        :return: *list* of info dicts
+        """
+        folder_list = convert_to_list(folder_list) if not isinstance(folder_list, list) else folder_list
+        if not self.silent:
+            self.__validate_file_list(folder_list)
+
+        cleaned_folder_list = []
+        for folder in folder_list:
+            folder = folder.replace("\\", "/")
+            folder = folder.rstrip("/")
+            folder += "/..."
+            cleaned_folder_list.append(folder)
+
+        changelist = self.__ensure_changelist(changelist)
+
+        info_dicts = self.run_cmd("reconcile", args=["-c", changelist], file_list=cleaned_folder_list)
+        return info_dicts
+
+    @validate_not_empty
     def delete_files(self, file_list, changelist="default"):
         """
         Marks files for delete
@@ -455,6 +555,9 @@ class P4Client(object):
         file_list = convert_to_list(file_list) if not isinstance(file_list, list) else file_list
         if not self.silent:
             self.__validate_file_list(file_list)
+
+        changelist = self.__ensure_changelist(changelist)
+
         info_dicts = self.run_cmd("delete", args=["-c", changelist], file_list=file_list)
         return info_dicts
 
@@ -465,9 +568,10 @@ class P4Client(object):
         :param perfect_match_only: bool* only delete if there's a perfect match of the changelist description
         :param case_sensitive: *bool*
         """
+        info_dicts = []
         cl_num = self.get_pending_changelists(changelist, perfect_match_only, case_sensitive)
         for cl in cl_num:
-            info_dicts = self.run_cmd('change', args=['-d', cl])
+            info_dicts.append(self.run_cmd('change', args=['-d', cl]))
             # TODO: Break down info dicts and look for errors
         return info_dicts
 
@@ -490,6 +594,7 @@ class P4Client(object):
 
         return files_and_cl
 
+    @validate_not_empty
     def add_or_edit_folders(self, folders, include_subfolders=True, changelist="default"):
         """
         Marks a folder for add or edit
@@ -513,6 +618,8 @@ class P4Client(object):
 
         return self.add_or_edit_files(all_files, changelist=changelist)
 
+
+    @validate_not_empty
     def add_or_edit_files(self, file_list, changelist="default"):
         """
         Marks the files in file_list for add if they are new, or edit if they are already versioned
@@ -561,6 +668,7 @@ class P4Client(object):
                 return int(self.__get_dict_value(info_dict, "change"))
         return -1
 
+    @validate_not_empty
     def edit_files(self, file_list, changelist="default"):
         """
         Marks the files in file_list for edit
@@ -578,9 +686,10 @@ class P4Client(object):
         info_dicts = self.run_cmd("edit", args=["-c", changelist], file_list=file_list)
         for info_dict in info_dicts:
             if self.__get_dict_value(info_dict, "code") == "error" and not self.silent:
-                print(self.__get_dict_value(info_dict, "data"))
+                logging.error(self.__get_dict_value(info_dict, "data"))
         return info_dicts
 
+    @validate_not_empty
     def add_files(self, file_list, changelist="default"):
         """
         Marks the files in file_list for add
@@ -604,15 +713,26 @@ class P4Client(object):
         :param changelist: *string* or *int* changelist number
         :return: *list* of depot paths
         """
+        changelist = self.__ensure_changelist(changelist)
         depot_paths = []
 
-        if type(changelist) == str:
-            try:
-                changelist = self.get_pending_changelists(changelist, perfect_match_only=True, case_sensitive=True)[0]
-            except IndexError as err:
-                return depot_paths
+        info_dicts = self.run_cmd("opened", args=["-c", changelist])
+        for info_dict in info_dicts:
+            for key, value in info_dict.items():
+                if "depotFile" in key.decode():
+                    depot_paths.append(value.decode())
 
-        info_dicts = self.run_cmd("describe", args=["-O", changelist])
+        return depot_paths
+
+    def get_all_files_in_all_changelists(self):
+        """
+        Does what it says on the box
+
+        :return: *list*
+        """
+        depot_paths = []
+
+        info_dicts = self.run_cmd("opened")
         for info_dict in info_dicts:
             for key, value in info_dict.items():
                 if "depotFile" in key.decode():
@@ -635,7 +755,13 @@ class P4Client(object):
 
         for info_dict in info_dicts:
             description_filter = description_filter.rstrip("\n")
-            cl_description = self.__get_dict_value(info_dict, "desc").rstrip("\n")
+            try:
+                cl_description = self.__get_dict_value(info_dict, "desc").rstrip("\n")
+            except AttributeError as err:
+                raise p4errors.P4cmdError(err)
+
+            if not cl_description:
+                logging.warning(f"The CL description is empty in this return object!\n{pformat(info_dict)}")
 
             if not case_sensitive:
                 description_filter = description_filter.lower()
@@ -654,9 +780,12 @@ class P4Client(object):
                         changelists.append([self.__get_dict_value(info_dict, "change"), cl_description])
 
         if descriptions:
-            return [pair[1] for pair in changelists]
+            return_list = [pair[1] for pair in changelists]
+            return_list.append("default")
         else:
-            return [int(pair[0]) for pair in changelists]
+            return_list = [int(pair[0]) for pair in changelists]
+            return_list.append("default")
+        return return_list
 
     def get_or_make_changelist(self, changelist_description, case_sensitive=False):
         """
@@ -678,7 +807,7 @@ class P4Client(object):
             case_sensitive=case_sensitive,
         )
 
-        if len(changelists):
+        if len(changelists) > 1: # > 1 because "default" is always appended
             return changelists[0]
         else:
             return self.make_new_changelist(description=changelist_description)
@@ -696,6 +825,7 @@ class P4Client(object):
 
         return workspaces
 
+    @validate_not_empty
     def get_depot_paths(self, paths):
         """
         Returns the depot path of the given files or folders, based on the client
@@ -715,6 +845,8 @@ class P4Client(object):
         depot_paths = [self.__get_dict_value(info, "depotFile").rstrip("/...") for info in info_dicts]
         return depot_paths
 
+
+    @validate_not_empty
     def get_local_paths(self, paths):
         """
         Returns the local path of the given files or folders, based on the client
@@ -735,6 +867,7 @@ class P4Client(object):
         local_paths = [self.__get_dict_value(info, "path") for info in info_dicts]
         return local_paths
 
+    @validate_not_empty
     def get_history(self, paths):
         """
         Returns changes for the given paths
@@ -763,6 +896,7 @@ class P4Client(object):
 
         try:
             sock = socket.create_connection((host, port), timeout=2)
+            sock.close()
             return True
         except:
             return False
@@ -827,7 +961,7 @@ class P4Client(object):
             except:
                 return dictionary.get(key, default_value)
 
-    def __fstat_to_p4_files(self, fstat_output_list, allow_invalid_files=False):
+    def fstat_to_p4_files(self, fstat_output_list, allow_invalid_files=False):
         """
         Turns the output of the fstat command into a list of P4File objects
 
@@ -837,6 +971,7 @@ class P4Client(object):
         :return:
         """
         p4files = []
+        p4_client = self.find_p4_client()
         for file_dict in fstat_output_list:
             p4file = P4File()
             p4file.set_depot_file_path(self.__get_dict_value(file_dict, "depotFile"))
@@ -854,9 +989,13 @@ class P4Client(object):
                 if other_open in key and key != other_open:
                     value = self.__get_dict_value(file_dict, key)
                     opened_by.append(value)
+                action_owner = "actionOwner" if sys.version_info == 2 else "actionOwner".encode()
+                if action_owner in key:
+                    value = self.__get_dict_value(file_dict, key)
+                    value = value.decode() + "@" + p4_client
+                    opened_by.append(value.encode())
 
             p4file.set_checked_out_by(opened_by)
-
             if allow_invalid_files:
                 p4files.append(p4file)
             else:
@@ -878,13 +1017,14 @@ class P4Client(object):
         current_dir = starting_dir
 
         while last_dir != current_dir:
-            for item in os.listdir(current_dir):
-                if item == ".p4config":
+            if os.path.exists(current_dir):
+                if ".p4config" in os.listdir(current_dir):
                     logging.info(".p4config found in %s" % current_dir)
                     self.perforce_root = current_dir
                     return True
+
             last_dir = current_dir
-            current_dir = os.path.abspath(current_dir + os.path.sep + os.pardir)
+            current_dir = os.path.dirname(last_dir)
         return False
 
     def __validate_file_list(self, file_list):
@@ -902,63 +1042,3 @@ class P4Client(object):
             if not f.lower().startswith(self.perforce_root.lower()):
                 raise Exception(f'{f} is not under perforce root: {self.perforce_root}')
 
-
-def split_list_into_strings_of_length(input_list, max_length=100):
-    """
-    From incoming list of strings, build joined together strings that are clamped to a max size
-
-    :param input_list:
-    :param max_length: str clamp length
-    """
-    full_list_str = " ".join([str(arg) for arg in input_list])
-
-    # if str is already below max length, just return it wrapped in a list
-    if not len(full_list_str) > max_length:
-        return [full_list_str]
-
-    # Keep appending on the last string until we hit the max length
-    # then add a new entry to the list and keep appending
-    clamped_str_list = [""]
-    for arg in input_list:
-        current_str = clamped_str_list[-1]
-        new_str = "{}{} ".format(current_str, arg)  # the space is deliberately placed, since the first item has no len
-
-        # if new string is too long, just append the arg as a new item and skip to next arg
-        if len(new_str) >= max_length:
-            clamped_str_list.append("{} ".format(arg))  #
-            continue
-
-        # set latest to updated string
-        clamped_str_list[-1] = new_str
-
-    return clamped_str_list
-
-
-def decode_dictionaries(info_dicts):
-    """
-    Decode list of dictionary keys and values into unicode from bytes
-    """
-    decoded_dicts = []
-    for info_dict in info_dicts:
-        decoded_dict = {k.decode(): v.decode() for k, v in info_dict.items()}
-        decoded_dicts.append(decoded_dict)
-    return decoded_dicts
-
-def convert_to_list(value):
-    if isinstance(value, tuple):
-        converted = [v for v in value]
-    else:
-        converted = [value]
-    return converted
-
-
-class HostOnline(object):
-    def __init__(self, host, port):
-        self.host = host
-        self.port = port
-
-    def __enter__(self):
-        pass
-
-    def __exit__(self, exc_type, exc_val, exc_tb):
-        pass
\ No newline at end of file
